/**
 * Inheritance Lock Protocol Header file
 * This is the API header use this, and only this, header file if you wish to use inheritance locks in your code.
 * Furthermore, you will need to link to the library during compilation.
 *
 * @author: W. David Frohlingsdorf
 * @date: 2016
 * @purpose: Dissertation, University of Glasgow
 * */
 
/// * * * * Necessary includes * * * *
#include <pthread.h>

/// * * * * Types * * * *
typedef void *INHERIT_LOCK;


/// * * * * Prototypes * * * *

//Reserves enough memory for a lock, initaliases the lock and returns a pointer to it.
//If address of a pointer is given as an argument (otherwise NULL), this pointer will also be directed 
//to the newly created lock. rec_lock allows a thread to acquire a lock more than once (nested)
//Returns NULL if an error occured
INHERIT_LOCK create_lock(INHERIT_LOCK*);
INHERIT_LOCK create_rec_lock(INHERIT_LOCK*);

//The calling thread tries to lock the given lock and might block until it is available.
//While the thread is blocked, previously acquired locks might be temporarly preempted according
//to the Protocol definition. Fails if thread tries to lock a non-recursive lock again.
//Returns 0 if successful, -1 otherwise
int acquire(INHERIT_LOCK);

//The calling thread release the given lock. Note that this also has to be the last lock that 
//has been acquired by this thread as required by the protocol. Returns 0 if successful -1 otherwise
//(illegal release of lock). If not successful, the thread still owns the lock. 
int release(INHERIT_LOCK);

//Frees the memory currently occupied by the lock object. The pointer to the lock, which address
//has to be given as an argument, will be set to NULL if successful. The function will fail if the
//lock is currently held by a thread or threads are waiting for it. Returns 0 if successful -1 otherwise.
int destroy_lock(INHERIT_LOCK*);
