/**
 * A generic map which hides the actual implementation of the 
 * mapping. However, the memory address of the key pointers 
 * will be used to check for equality. 
 *
 * @author: W. David Frohlingsdorf
 * @date: 2016
 * @purpose: Dissertation, University of Glasgow
 * */

#include <stddef.h>

//Gets the value associated with key
//if no value is associated with key NULL is returned
//@param Pointer to the key
//@return Pointer to the value
void *map_get(void *key);

//Adds a tuple to the map
//Both key and value have to point
//to a valid memory address. In particular
//NULL is not allowed as it would disrupt
//the functionality. 
//@param Pointer to the key
//@param Pointer to the value
void map_add(void *key, void *value);

//Deletes an entry in the map
//If no value is associated with
//the key, this function call has no
//effect. 
//The memory pointed to by key or value
//will not be freed or effected by this
//function
//@param Pointer to the key
//@return Pointer to the value (NULL if key
//        is not in the map)
void *map_del(void *key);
