/**
 * Inheritance Lock Protocol Artificial Deadlock test
 *
 * @author: W. David Frohlingsdorf
 * @date: 2016
 * @purpose: Dissertation, University of Glasgow
 * */

#include <pthread.h>
#include <unistd.h>
#include <stdio.h>

#include "inheritance_lock_api.h"

#define SLEEPTIME 2
#define TIMEOUT   10

int x, y;
char turn;
volatile char running;
INHERIT_LOCK lockX, lockY;

void *run1(void *param) /*Competing for x*/
{
  while(running)
  {
    acquire(lockX);
    {
      x = 1;
      turn = 2;
      while(turn != 1)
        x = 1;
      x = 1;
    }
    release(lockX);
  }
  printf("run1() done\n");
  return NULL;
}

void *run2(void *param) /*Competing for y*/
{
  while(running)
  {
    acquire(lockY);
    {
      y = 2;
      turn = 1;
      while(turn != 2)
        y = 2;
      y = 2;
    }
    release(lockY);
  }
  printf("run2() done\n");
  return NULL;
}

void *run3(void *param) /*Starved, needs x and y*/
{
  acquire(lockX);
  acquire(lockY);
  {
    x = 3;
    y = 3;
  }
  release(lockY);
  release(lockX);
  running = 0;
  printf("run3() done\n");
  return NULL;
}

int main()
{
  pthread_t t1, t2, t3;
  turn = 1;
  running = 1;
  create_lock(&lockX);
  create_lock(&lockY);
  pthread_create(&t1, NULL, run1, NULL);
  pthread_create(&t2, NULL, run2, NULL);
  sleep(SLEEPTIME);
  pthread_create(&t3, NULL, run3, NULL);
  printf("Waiting for result...\n");
  sleep(TIMEOUT);
  if(running == 0)
  {
    printf("Not starved\n");
    pthread_join(t1, NULL);
    pthread_join(t2, NULL);
    pthread_join(t3, NULL);
  }
  else
    printf("Timed out. Starvation!\n");
  printf("Final results:\n\tx = %d\n\ty = %d\n", x, y);
  return 0;
}
  
