/**
 * Simple tests for the Inheritance Lock protocol
 *
 * @author: W. David Frohlingsdorf
 * @date: 2016
 * @purpose: Dissertation, University of Glasgow
 * */

#include <unistd.h>  /* For sleep() */
#include <stdio.h>
#include "inheritance_lock_api.h"

INHERIT_LOCK lock1;
INHERIT_LOCK lock2;

void *run(void *n)
{
    printf("Run: I've been created. Acquire lock2 now.\n");
    acquire(lock2);
    printf("Run: That's that done! Now lock1.\n");
    acquire(lock1);
    printf("Run: Got both, will now release lock1 and 2.\n");
    release(lock1);
    release(lock2);
    printf("Run: All done!\n");
    return NULL;
}

void *infinite_run()
{
    printf("Infinite Run: I've been created.\n");
    acquire(lock1);
    acquire(lock2);
    printf("Infinite Run: Acquired locks, busy wait now.\n");
    while(1)
        sleep(1);//Busy wait
    printf("Infinite Run: You should never see me.\n");
    return NULL;
}

void *busy_run()
{
    int i;
    printf("Busy Run: I've been created.\n");
    acquire(lock1);
    for(i=0; i<10; i++)
    {
        printf("Busy Run: I am terrible busy today (%d).\n", i);
        sleep(1);
    }
    printf("Busy Run: I am exhausted.\n");
    release(lock1);
    printf("Busy Run: That's me done.\n");
    return NULL;
}

int main()
{
    pthread_t t;
    create_lock(&lock1);
    create_lock(&lock2);

    acquire(lock1);
    pthread_create(&t, NULL, run, NULL);
    printf("Main: Created thread. Going to sleep...\n");
    sleep(3);
    printf("Main: Finished sleeping. Release lock1 now.\n");
    release(lock1);
    printf("Main: Released lock1. Now joining with thread...\n");
    pthread_join(t, NULL);
    printf("Main: Joined with thread.\n");

    printf("Main: Create infinite thread.\n");
    pthread_create(&t, NULL, infinite_run, NULL);
    printf("Main: Created, going to sleep now..\n");
    sleep(3);
    printf("Main: Finished sleeping. KILL THIS THREAD!!\n");
    pthread_cancel(t);
    printf("Main: Cancelled the thread, let's try joining.\n");
    pthread_join(t, NULL);
    printf("Main: Joined, let's get those locks.\n");
    acquire(lock1);
    acquire(lock2);
    printf("Main: Done, now releasing.\n");
    release(lock2);
    release(lock1);
    printf("Main: Both released.\n");

    pthread_create(&t, NULL, busy_run, NULL);
    printf("Main: Busy run has been created. Let's give it a bit time.\n");
    sleep(3);
    printf("Main: Now trying to get lock1.\n");
    acquire(lock1);
    printf("Main: I've got it!.\n");
    release(lock1);
    printf("Main: ..and released it.\n");
    pthread_join(t, NULL);
    printf("Main: All done, let's call it a day!\n");
    
    return 0;
}
