/**
 * Dynamic shared library which when loaded
 * replaces all pthread mutecies calls to
 * a stub function. Please check the readme
 * file in this directory for further details.
 *
 * Uses a generic map as defined in map.h 
 *
 * @author: W. David Frohlingsdorf
 * @date: 2017
 * @purpose: Dissertation, University of Glasgow
 * */

/// * * * Includes * * *

#define _GNU_SOURCE

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <bits/pthreadtypes.h>
#include <dlfcn.h>

/// * * * Definitions and Macros * * *

#define VERBOSE 0     //0=Off, 1=On
#define GLIBC_VERSION "GLIBC_2.3.2"

/// * * * Redirected functions * * *

#undef pthread_mutex_lock

int pthread_mutex_lock(pthread_mutex_t *mutex)
{
  if(VERBOSE)
    printf("Caught pthread_mutex_lock()\n");
  return 0;
}

#undef pthread_mutex_unlock

int pthread_mutex_unlock(pthread_mutex_t *mutex)
{
  if(VERBOSE)
    printf("Caught pthread_mutex_unlock()\n");
  return 0;
}

#undef pthread_mutex_trylock

int pthread_mutex_trylock(pthread_mutex_t *mutex)
{
  if(VERBOSE)
    printf("Caught pthread_mutex_trylock()\n");
  return 0;  
}

#undef pthread_mutex_destroy

int pthread_mutex_destroy(pthread_mutex_t *mutex)
{
  static int (*real_destroy)(pthread_mutex_t *) = NULL;
  if(!real_destroy)
    real_destroy = dlsym(RTLD_NEXT, "pthread_mutex_destroy");
  if(VERBOSE)
    printf("Caught pthread_mutex_destroy()\n");
  return real_destroy(mutex);
}

