-- Author: W. David Frohlingsdorf
-- Date: 2016
-- Purpose: Dissertation, Univeristy of Glasgow
--
-- This program creates a binary which takes two
-- arguments, number of threads and number of
-- locks, and outputs a promela source file
-- on standard output.

import System.Exit
import System.Environment
import Data.Set (Set)
import qualified Data.Set as Set

main :: IO ()
main = do
    args <- getArgs
    let (threads, locks) = evalArgs args
    printHeader
    putStr "#define N "
    printVar threads
    putStr  " /* Number of threads */\n"
    putStr "#define L "
    printVar locks
    putStr  " /* Number of locks */\n"
    printPreBody
    printBody (Set.fromList [1 .. locks]) 0 2
    printPostBody
    
evalArgs :: [String] -> (Int,Int)
evalArgs [a, b] = (read a::Int, read b::Int)
evalArgs _ = error "Usage: creator.exe NumberOfThreads NumberOfLocks"

printVar :: Int -> IO ()
printVar x = putStr $ show x

printBody :: Set Int -> Int -> Int -> IO ()
printBody set cur space 
    | Set.size set == 0 = putStr ""
    | otherwise = 
        do
        putStr $ replicate space ' '
        if (cur==0) then putStr "do\n" else putStr "if\n"
        mapM_ (\x -> handleRec x set space) set 
        putStr $ replicate space ' '
        if (cur==0) then putStr "od;\n" 
        else putStr (":: skip\n"++(replicate space ' ')++"fi;\n"++(replicate space ' ')++"m.lock = "++(show cur)++";"++(if cur == 1 then " critical++; critical--;" else "")++"\n")
        
handleRec :: Int -> Set Int -> Int -> IO ()
handleRec x set space = do
    putStr $ replicate space ' '
    putStr "::  m.lock = "
    printVar x
    putStr ";\n"
    putStr $ replicate space ' '
    putStr "    acquire!m; acknowledge[id]?1;"
    if x == 1
    then putStr " critical++; critical--;\n"
    else putStr "\n"
    printBody (Set.delete x set) x (space+4)
    putStr $ replicate space ' '
    putStr "    release!m; acknowledge[id]?1;\n"
    
    
printHeader :: IO ()
printHeader =
    putStr
    "/*\n\
    \ * author: W. David Frohlingsdorf\n\
    \ * date: 2016\n\
    \ * purpose: Dissertation, University of Glasgow\n\
    \ *\n\
    \ * This is a model of the inheritance lock protocol\n\
    \ * for the SPIN model checker.\n\
    \ * SPIN verifies that neither starvation nor deadlock\n\
    \ * is possible with the inheritance lock protcol.\n\
    \ * Moreover SPIN checks that mutual exclusion is preserved.\n\
    \ *\n\
    \ * Inheritance locks are non-recursive\n\
    \ * No further checks are done regarding wheather a request is valid or not\n\
    \ */\n\
    \\n"
    
printPreBody :: IO ()
printPreBody =
    putStr
    "\n\
    \typedef message\n\
    \{\n\
    \  short thrd;\n\
    \  short lock\n\
    \};\n\
    \\n\
    \typedef waitGraph\n\
    \{\n\
    \  short q[N]\n\
    \};\n\
    \\n\
    \chan acquire = [N] of {message}\n\
    \chan release = [N] of {message}\n\
    \chan acknowledge[N+1] = [1] of {bit} /*acknowledge[0] not used*/\n\
    \short critical = 0 /*Add 1 if in critical region of lock 1. Used for LTL property*/\n\
    \short p1 = -1;\n\
    \\n\
    \proctype thread(short id)\n\
    \{\n\
    \  message m;\n\
    \  m.thrd = id;\n\
    \  label_thread:\n"
    
printPostBody :: IO ()
printPostBody =
    putStr
    "}\n\
    \\n\
    \proctype controller()\n\
    \{\n\
    \  message m;\n\
    \  waitGraph heldBy[L+1];\n\
    \  short waitsFor[N+1];\n\
    \  short root;\n\
    \  short lock;\n\
    \  short i;\n\
    \  do\n\
    \  ::  acquire?m;\n\
    \      atomic\n\
    \      {\n\
    \        if\n\
    \        :: (heldBy[m.lock].q[0] == 0) -> root = 0\n\
    \        :: else ->\n\
    \            lock = m.lock;\n\
    \            do\n\
    \            :: (lock == 0) -> break\n\
    \            :: else -> root = heldBy[lock].q[0]; lock = waitsFor[root]\n\
    \            od;\n\
    \        fi;\n\
    \      }\n\
    \      if\n\
    \      :: (root == 0) -> heldBy[m.lock].q[0] = m.thrd; acknowledge[m.thrd]!1\n\
    \      :: (root == m.thrd) -> acknowledge[m.thrd]!1;\n\
    \      :: else -> atomic{\n\
    \            waitsFor[m.thrd] = m.lock;\n\
    \            i = 1;\n\
    \            do\n\
    \            :: (heldBy[m.lock].q[i] == 0) -> heldBy[m.lock].q[i] = m.thrd; break\n\
    \            :: else -> i = i + 1;\n\
    \            od;\n\
    \          }\n\
    \      fi;\n\
    \\n\
    \\n\
    \  :: release?m;\n\
    \      if\n\
    \      :: (heldBy[m.lock].q[0] != m.thrd) -> acknowledge[m.thrd]!1        /*Inherited this lock*/\n\
    \      :: (heldBy[m.lock].q[0] == m.thrd && heldBy[m.lock].q[1] == 0) ->  /*Owner of the lock, but no one is waiting*/\n\
    \          heldBy[m.lock].q[0] = 0; acknowledge[m.thrd]!1\n\
    \      :: else -> atomic{                                                 /*Owner and someone is waiting*/\n\
    \            i = 0;\n\
    \            do\n\
    \            :: (i < N-1) -> heldBy[m.lock].q[i] = heldBy[m.lock].q[i+1]; i = i + 1 \n\
    \            :: else -> heldBy[m.lock].q[N-1] = 0; break\n\
    \            od;\n\
    \            waitsFor[heldBy[m.lock].q[0]] = 0;\n\
    \          }\n\
    \          acknowledge[heldBy[m.lock].q[0]]!1;\n\
    \          acknowledge[m.thrd]!1;\n\
    \      fi; \n\
    \  od;\n\
    \}\n\
    \\n\
    \init\n\
    \{\n\
    \  int i;\n\
    \  atomic\n\
    \  {\n\
    \    run controller();\n\
    \    for (i: 1 .. N){\n\
    \      p1 = run thread(i);\n\
    \    }\n\
    \  }\n\
    \}\n\
    \\n\
    \#define p (thread[p1]@label_thread)\n\
    \#define q (critical == 0 || critical == 1)\n\
    \\n\
    \ltl property1 {[]((<>p) && q)}\n\
    \\n"
