/*
 * author: W. David Frohlingsdorf
 * date: 2016
 * purpose: Dissertation, University of Glasgow
 *
 * This is a model of the inheritance lock protocol
 * for the SPIN model checker.
 * SPIN verifies that neither starvation nor deadlock
 * is possible with the inheritance lock protcol.
 * Moreover SPIN checks that mutual exclusion is preserved.
 *
 * Inheritance locks are non-recursive
 * No further checks are done regarding wheather a request is valid or not
 */

#define N 3 /* Number of threads */
#define L 3 /* Number of locks */

typedef message
{
  short thrd;
  short lock
};

typedef waitGraph
{
  short q[N]
};

chan acquire = [N] of {message}
chan release = [N] of {message}
chan acknowledge[N+1] = [1] of {bit} /*acknowledge[0] not used*/
short critical = 0 /*Add 1 if in critical region of lock 1. Used for LTL property*/
short p1 = -1;

proctype thread(short id)
{
  message m;
  m.thrd = id;
  label_thread:
  do
  ::  m.lock = 1;
      acquire!m; acknowledge[id]?1; critical++; critical--;
      if
      ::  m.lock = 2;
          acquire!m; acknowledge[id]?1;
          if
          ::  m.lock = 3;
              acquire!m; acknowledge[id]?1;
              release!m; acknowledge[id]?1;
          :: skip
          fi;
          m.lock = 2;
          release!m; acknowledge[id]?1;
      ::  m.lock = 3;
          acquire!m; acknowledge[id]?1;
          if
          ::  m.lock = 2;
              acquire!m; acknowledge[id]?1;
              release!m; acknowledge[id]?1;
          :: skip
          fi;
          m.lock = 3;
          release!m; acknowledge[id]?1;
      :: skip
      fi;
      m.lock = 1; critical++; critical--;
      release!m; acknowledge[id]?1;
  ::  m.lock = 2;
      acquire!m; acknowledge[id]?1;
      if
      ::  m.lock = 1;
          acquire!m; acknowledge[id]?1; critical++; critical--;
          if
          ::  m.lock = 3;
              acquire!m; acknowledge[id]?1;
              release!m; acknowledge[id]?1;
          :: skip
          fi;
          m.lock = 1; critical++; critical--;
          release!m; acknowledge[id]?1;
      ::  m.lock = 3;
          acquire!m; acknowledge[id]?1;
          if
          ::  m.lock = 1;
              acquire!m; acknowledge[id]?1; critical++; critical--;
              release!m; acknowledge[id]?1;
          :: skip
          fi;
          m.lock = 3;
          release!m; acknowledge[id]?1;
      :: skip
      fi;
      m.lock = 2;
      release!m; acknowledge[id]?1;
  ::  m.lock = 3;
      acquire!m; acknowledge[id]?1;
      if
      ::  m.lock = 1;
          acquire!m; acknowledge[id]?1; critical++; critical--;
          if
          ::  m.lock = 2;
              acquire!m; acknowledge[id]?1;
              release!m; acknowledge[id]?1;
          :: skip
          fi;
          m.lock = 1; critical++; critical--;
          release!m; acknowledge[id]?1;
      ::  m.lock = 2;
          acquire!m; acknowledge[id]?1;
          if
          ::  m.lock = 1;
              acquire!m; acknowledge[id]?1; critical++; critical--;
              release!m; acknowledge[id]?1;
          :: skip
          fi;
          m.lock = 2;
          release!m; acknowledge[id]?1;
      :: skip
      fi;
      m.lock = 3;
      release!m; acknowledge[id]?1;
  od;
}

proctype controller()
{
  message m;
  waitGraph heldBy[L+1];
  short waitsFor[N+1];
  short root;
  short lock;
  short i;
  do
  ::  acquire?m;
      atomic
      {
        if
        :: (heldBy[m.lock].q[0] == 0) -> root = 0
        :: else ->
            lock = m.lock;
            do
            :: (lock == 0) -> break
            :: else -> root = heldBy[lock].q[0]; lock = waitsFor[root]
            od;
        fi;
      }
      if
      :: (root == 0) -> heldBy[m.lock].q[0] = m.thrd; acknowledge[m.thrd]!1
      :: (root == m.thrd) -> acknowledge[m.thrd]!1;
      :: else -> atomic{
            waitsFor[m.thrd] = m.lock;
            i = 1;
            do
            :: (heldBy[m.lock].q[i] == 0) -> heldBy[m.lock].q[i] = m.thrd; break
            :: else -> i = i + 1;
            od;
          }
      fi;


  :: release?m;
      if
      :: (heldBy[m.lock].q[0] != m.thrd) -> acknowledge[m.thrd]!1        /*Inherited this lock*/
      :: (heldBy[m.lock].q[0] == m.thrd && heldBy[m.lock].q[1] == 0) ->  /*Owner of the lock, but no one is waiting*/
          heldBy[m.lock].q[0] = 0; acknowledge[m.thrd]!1
      :: else -> atomic{                                                 /*Owner and someone is waiting*/
            i = 0;
            do
            :: (i < N-1) -> heldBy[m.lock].q[i] = heldBy[m.lock].q[i+1]; i = i + 1 
            :: else -> heldBy[m.lock].q[N-1] = 0; break
            od;
            waitsFor[heldBy[m.lock].q[0]] = 0;
          }
          acknowledge[heldBy[m.lock].q[0]]!1;
          acknowledge[m.thrd]!1;
      fi; 
  od;
}

init
{
  int i;
  atomic
  {
    run controller();
    for (i: 1 .. N){
      p1 = run thread(i);
    }
  }
}

#define p (thread[p1]@label_thread)
#define q (critical == 0 || critical == 1)

ltl property1 {[]((<>p) && q)}

